package pps.core.system.service;


import cn.hutool.core.text.CharSequenceUtil;
import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import org.apache.commons.lang3.StringUtils;
import pps.core.common.session.PpsUserSession;
import pps.core.common.utils.PatternUtil;
import pps.core.common.utils.VerificationCode;
import pps.core.system.cache.LoginSmsCodeCache;
import pps.core.system.cache.LoginVerCodeCache;
import pps.core.system.config.LoginConfig;
import pps.core.system.entity.SysUserEnt;
import pps.core.system.entity.SysUserForgetLogEnt;
import pps.core.system.entity.SysUserLoginLogEnt;
import pps.core.system.entity.SysUserOrganizationRelEnt;
import pps.core.system.error.LoginError;
import pps.core.system.mapper.SysUserForgetLogMapper;
import pps.core.system.mapper.SysUserLoginLogMapper;
import pps.core.system.mapper.SysUserMapper;
import pps.core.system.mapper.SysUserOrganizationRelMapper;
import pps.core.system.service.data.sys_login.*;
import pps.core.system.utils.UserInfoUtil;
import xstartup.annotation.XService;
import xstartup.annotation.XText;
import xstartup.base.XContext;
import xstartup.base.XSsoAuth;
import xstartup.base.XUser;
import xstartup.base.conf.XConfManager;
import xstartup.base.exception.XServiceException;
import xstartup.base.util.XDateUtils;
import xstartup.base.util.XStringUtils;
import xstartup.base.util.XUuidUtils;
import xstartup.core.base.provider.data.XToken;
import xstartup.data.XFileResult;
import xstartup.data.XServiceResult;
import xstartup.data.XSingleResult;
import xstartup.feature.api.annotation.XApiGet;
import xstartup.feature.api.annotation.XApiPost;
import xstartup.feature.api.conf.XApiCookieConf;
import xstartup.helper.XTokenHelper;

import java.awt.image.BufferedImage;
import java.util.Date;
import java.util.Map;
import java.util.UUID;

import static pps.core.common.constant.LoginConstant.LOGIN_FLAG_SIMULATE;
import static pps.core.system.error.LoginError.NotExistUserName;
import static pps.core.system.error.LoginError.Trade_IpWhiteError;

@XService
public class LoginService {

    @XText("账号密码登录")
    @XApiPost(anonymous = true)
    public XSingleResult<LoginByAccountOutput> loginByAccount(XContext context, LoginByAccountInput input) {
        Long count = getErrorTimes(context, input.getLoginName(), null);
        if (count >= 5) {
            return XSingleResult.error(context, LoginError.ISLOCKED);
        }
        checkCode(context, input.getCode(), input.getCodeIden());
        XSingleResult<SysUserEnt> sysUserResult = checkAccount(context, input.getLoginName(), input.getPassword());
        SysUserEnt sysUserEnt = sysUserResult.getResult();
        if (sysUserEnt.getStatus() != null && sysUserEnt.getStatus() == 1)
            return XSingleResult.error(context, LoginError.ISLOCKED);
        long getTime = System.currentTimeMillis();
        if (sysUserEnt.getIsOuter() != null && sysUserEnt.getIsOuter() != 1)
            return XSingleResult.error(context, LoginError.ISOUT);
        Date now = new Date();
        SysUserOrganizationRelMapper relMapper = context.getBean(SysUserOrganizationRelMapper.class);
        QueryWrapper<SysUserOrganizationRelEnt> relQueryWrapper = new QueryWrapper<>();
        relQueryWrapper.lambda().eq(SysUserOrganizationRelEnt::getUserId, sysUserEnt.getId());
        relQueryWrapper.lambda().le(SysUserOrganizationRelEnt::getEffectTime, now);
        relQueryWrapper.lambda().gt(SysUserOrganizationRelEnt::getEndTime, now);
        SysUserOrganizationRelEnt sysUserOrganizationRelEnt = relMapper.selectOne(relQueryWrapper);
        if (sysUserOrganizationRelEnt == null) {
            context.getLogger().warn("用户未挂靠组织");
            sysUserOrganizationRelEnt = new SysUserOrganizationRelEnt();
            sysUserOrganizationRelEnt.setOuId("1");
        }
        XApiCookieConf apiCookieConf = XConfManager.getConf(XApiCookieConf.class);
        Integer expiresIn = apiCookieConf.getExpires();
        String userName = sysUserEnt.getUserName(); //姓名
        Long orgId = 1L; //组织机构id
        Long roleId0 = 1L; //角色id1
        Long roleId1 = 1L; //角色id2
        Long userId = IdWorker.getId(); //用户id
        String userCode = sysUserEnt.getId();//这个是用户字符串类型的唯一标志
        String ssoToken = null; //sso系统授权的token
        String ssoRefreshToken = null; //sso系统授权的刷新token
        String uniqueId = XUuidUtils.randomUUID();
        XUser user = XUser.createUser(XUser.ENTRY_API,
                orgId,
                new Long[]{roleId0, roleId1},
                userId,
                userCode,
                sysUserEnt.getLoginName(),
                userName,
                new XSsoAuth(uniqueId, ssoToken, ssoRefreshToken));
        XSingleResult<XToken> saveUserResult = XTokenHelper.saveUser(context, user, expiresIn);
        saveUserResult.throwIfFail();
        LoginByAccountOutput output = new LoginByAccountOutput();
        output.setExpiresIn(expiresIn);
        output.setToken(saveUserResult.getResult().getToken());
        PpsUserSession session = context.getSession(PpsUserSession.class);
        output.setSession(session);
        //登录成功后，删除失败记录
        deleteError(context, sysUserEnt);
        return XSingleResult.success(output);
    }

    @XText("忘记密码")
    @XApiPost(anonymous = true)
    public XSingleResult<ForgetPwdOutput> forgetPwd(XContext context, ForgetPwdInput input) {
        if (XStringUtils.isNotEmpty(input.getCode()))
            checkCode(context, input.getCode(), input.getCodeIden());
        Long count = getForgetErrorTimes(context);
        if (count >= 5) {
            return XSingleResult.error(500, "查询失败超过5次");
        }
        XSingleResult<SysUserEnt> sysUserResult = checkForgetAccount(context, input.getLoginName());
        SysUserEnt sysUserEnt = sysUserResult.getResult();
        String phone = sysUserEnt.getPhone();
        phone = UserInfoUtil.decrypt(phone);
        if (StringUtils.isBlank(phone)) {
            return XSingleResult.error(context, LoginError.NotExistPhone);
        }
        if (!PatternUtil.checkPhone(phone)) {
            return XSingleResult.error(context, LoginError.NotExistPhone);
        }
        ForgetPwdOutput output = new ForgetPwdOutput();
        output.setPhone(phone.substring(phone.length() - 4));
        return XSingleResult.success(output);
    }

    @XText("修改密码")
    @XApiPost(anonymous = true)
    public XServiceResult modifyPwd(XContext context, ModifyPwdInput input) {
        checkSmsCode(context, input.getLoginName() + "-pwd", input.getSmsCode());
        XSingleResult<SysUserEnt> sysUserResult = checkAccount(context, input.getLoginName(), null);
        SysUserEnt sysUserEnt = sysUserResult.getResult();
        sysUserEnt.setPassword(input.getPassword());
        SysUserMapper userMapper = context.getBean(SysUserMapper.class);
        userMapper.updateById(sysUserEnt);
        return XServiceResult.OK;
    }

    /**
     * 通过模拟登录
     *
     * @param context 上下文
     * @param input   输入
     * @return {@link XSingleResult}<{@link LoginBySimulateOutput}>
     */
    @XText("模拟用户登陆")
    @XApiGet(anonymous = true)
    public XSingleResult<LoginBySimulateOutput> thirdToken(XContext context, LoginBySimulateInput input) {
        if (!CharSequenceUtil.equals(input.getAuthKey(), "66b500fcaff049dca77f66733d9bb172")) {
            return XSingleResult.error(context, Trade_IpWhiteError);
        }
        SysUserMapper mapper = context.getBean(SysUserMapper.class);
        SysUserEnt sysUserEnt = mapper.selectOne(
                new LambdaQueryWrapper<SysUserEnt>()
                        .eq(SysUserEnt::getLoginName, "cqyanshi")
        );
        if (ObjectUtil.isNull(sysUserEnt)) {
            return XSingleResult.error(context, NotExistUserName);
        }
        Integer expiresIn = XConfManager.getConf(XApiCookieConf.class).getExpires();
        XSingleResult<XToken> saveUserResult = XTokenHelper.saveUser(context,
                XUser.createUser(XUser.ENTRY_API,
                        LOGIN_FLAG_SIMULATE,
                        new Long[]{1L, 1L},
                        IdWorker.getId(),
                        sysUserEnt.getId(),
                        "cqyanshi",
                        sysUserEnt.getUserName(),
                        new XSsoAuth(XUuidUtils.randomUUID(), null, null))
                , expiresIn);
        saveUserResult.throwIfFail();
        return XSingleResult.success(LoginBySimulateOutput.builder()
                .expiresIn(expiresIn)
                .token(saveUserResult.getResult().getToken())
                .build());
    }

    @XText("退出")
    @XApiPost
    public XServiceResult logout(XContext context) {
        XUser user = context.getUser();
        XTokenHelper.deleteUser(context, user);
        context.updateToken(context.getWebRequest().getToken(), 0);
        return XServiceResult.OK;
    }

    @XText("验证码")
    @XApiGet(anonymous = true)
    public XFileResult getVerificationCode(XContext context, VerificationCodeInput input) {
        Map<String, Object> codeMap = VerificationCode.generateCodeAndPic();
        // 将四位数字的验证码保存到Session中。
        LoginVerCodeCache cache = new LoginVerCodeCache();
        cache.setCodeKey(input.getCodeIden());
        cache.setCode(codeMap.get("code").toString());
        LoginVerCodeCache.set(context, cache);
        return XFileResult.createWithPng(input.getCodeIden() + ".png", (BufferedImage) codeMap.get("codePic"));
    }

    /*------------------------------- private -------------------------------*/

    private boolean checkIsSendSmsCode(XContext context) {
        Integer sendSmsCodeSwitch = new LoginConfig().sendSmsCodeSwitch.findValue(context);
        if (sendSmsCodeSwitch != null && sendSmsCodeSwitch == 1) {
            return true;
        }
        return false;
    }

    private XSingleResult<SysUserEnt> checkAccount(XContext context, String loginName, String password) {
        SysUserMapper mapper = context.getBean(SysUserMapper.class);
        QueryWrapper<SysUserEnt> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(SysUserEnt::getLoginName, loginName);
        SysUserEnt sysUserEnt = mapper.selectOne(queryWrapper);
        if (sysUserEnt == null) {
            insertError(context, loginName);
            Long count = getErrorTimes(context, loginName, sysUserEnt);
            throw new XServiceException(1000, "用户名或密码错误，还剩" + (5 - count) + "次机会");
        }
        if (StringUtils.isNotBlank(password)) {
            String decrypt = sysUserEnt.getPassword();
            if (!password.equals(decrypt)) {
                insertError(context, loginName);
                Long count = getErrorTimes(context, loginName, sysUserEnt);
                throw new XServiceException(1000, "用户名或密码错误，还剩" + (5 - count) + "次机会");
            }
        }
        return XSingleResult.success(sysUserEnt);
    }

    private void checkSmsCode(XContext context, String userKey, String smsCode) {
        if (!checkIsSendSmsCode(context)) {
            return;
        }
        if (StringUtils.isBlank(smsCode)) {
            throw new XServiceException(context, LoginError.NotValidateSmsCode);
        }
        LoginSmsCodeCache smsCodeCache = LoginSmsCodeCache.exist(context, userKey);
        if (smsCodeCache != null) {
            String code = smsCodeCache.getCode();
            if (code.equals(smsCode)) {
                LoginSmsCodeCache.delete(context, userKey);
                return;
            }
        }
        throw new XServiceException(context, LoginError.NotValidateSmsCode);
    }

    private void checkCode(XContext context, String code, String codeIden) {
        if (code == null) {
            throw new XServiceException(context, LoginError.NotExistCode);
        }
        LoginVerCodeCache isUse = LoginVerCodeCache.exist(context, codeIden);
        if (isUse == null) {
            throw new XServiceException(context, LoginError.NotUsedCode);
        }
        if (!isUse.getCode().toLowerCase().equals(code.toLowerCase())) {
            throw new XServiceException(context, LoginError.NotCompareCode);
        }
    }

    private void deleteError(XContext context, SysUserEnt sysUserEnt) {
        //删除登录日志错误信息
        SysUserLoginLogMapper userLoginLogMapper = context.getBean(SysUserLoginLogMapper.class);
        QueryWrapper<SysUserLoginLogEnt> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(SysUserLoginLogEnt::getLoginName, sysUserEnt.getLoginName());
        userLoginLogMapper.delete(queryWrapper);
        SysUserMapper mapper = context.getBean(SysUserMapper.class);
        sysUserEnt.setStatus(0);
        sysUserEnt.setModifyTime(new Date());
        mapper.updateById(sysUserEnt);
    }

    private void insertError(XContext context, String loginName) {
        //记录登录日志错误信息
        SysUserLoginLogMapper userLoginLogMapper = context.getBean(SysUserLoginLogMapper.class);
        SysUserLoginLogEnt logEnt = new SysUserLoginLogEnt();
        logEnt.setId(UUID.randomUUID().toString());
        logEnt.setErrorMsg("账户、密码错误");
        logEnt.setLoginName(loginName);
        logEnt.setCreateTime(new Date());
        userLoginLogMapper.insert(logEnt);
    }

    private XSingleResult<SysUserEnt> checkForgetAccount(XContext context, String loginName) {
        SysUserMapper mapper = context.getBean(SysUserMapper.class);
        QueryWrapper<SysUserEnt> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(SysUserEnt::getLoginName, loginName);
        SysUserEnt sysUserEnt = mapper.selectOne(queryWrapper);
        if (sysUserEnt == null) {
            insertForgetError(context, loginName);
            Long count = getForgetErrorTimes(context);
            throw new XServiceException(1000, "用户名错误，还剩" + (5 - count) + "次机会");
        }
        return XSingleResult.success(sysUserEnt);
    }

    private void insertForgetError(XContext context, String loginName) {
        //记录登录日志错误信息
        SysUserForgetLogMapper userForgetLogMapper = context.getBean(SysUserForgetLogMapper.class);
        SysUserForgetLogEnt logEnt = new SysUserForgetLogEnt();
        logEnt.setId(UUID.randomUUID().toString());
        logEnt.setIpAddress(context.getWebRequest().getClientIp());
        logEnt.setLoginName(loginName);
        logEnt.setCreateTime(new Date());
        userForgetLogMapper.insert(logEnt);
    }

    private Long getErrorTimes(XContext context, String loginName, SysUserEnt sysUserEnt) {
        SysUserLoginLogMapper userLoginLogMapper = context.getBean(SysUserLoginLogMapper.class);
        QueryWrapper<SysUserLoginLogEnt> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(SysUserLoginLogEnt::getLoginName, loginName);
        queryWrapper.lambda().ge(SysUserLoginLogEnt::getCreateTime, XDateUtils.getDayBeginTime(new Date()));
        Long count = userLoginLogMapper.selectCount(queryWrapper);
        if (count >= 5 && sysUserEnt != null) {
            SysUserMapper mapper = context.getBean(SysUserMapper.class);
            sysUserEnt.setStatus(2);
            sysUserEnt.setModifyTime(new Date());
            mapper.updateById(sysUserEnt);
        }
        return count;
    }

    private Long getForgetErrorTimes(XContext context) {
        SysUserForgetLogMapper userLoginLogMapper = context.getBean(SysUserForgetLogMapper.class);
        QueryWrapper<SysUserForgetLogEnt> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(SysUserForgetLogEnt::getIpAddress, context.getWebRequest().getClientIp());
        queryWrapper.lambda().ge(SysUserForgetLogEnt::getCreateTime, XDateUtils.addMunites(new Date(), -5));
        Long count = userLoginLogMapper.selectCount(queryWrapper);
        return count;
    }
}